from .feature_state import FeatureState
from .state_repository import StateRepository


class FeatureSnapshot(object):
    def __init__(self, state_repository: StateRepository):
        self.state_repository = state_repository
        self.feature_states = self.state_repository.get_feature_states()

    def get_feature_names(self) -> list:
        return [k for k in self.feature_states.keys()]

    def get_feature_state(self, name: str) -> FeatureState:
        feature_state = self.feature_states.get(name, None) \
                        or self.state_repository.get_feature_state(name)
        if feature_state:
            self.feature_states[name] = feature_state
        return feature_state

    def set_feature_state(self, feature_state: FeatureState) -> None:
        self.state_repository.set_feature_state(feature_state)
        self.feature_states[feature_state.name] = feature_state



from copy import copy
from copy import deepcopy

from .feature_state import FeatureState


class StateRepository(object):
    def __init__(self):
        self.feature_states = self.load()

    def load(self) -> dict:
        """to be implemented"""
        return {}

    def save(self, feature_state: FeatureState):
        """to be implemented"""
        pass

    def get_feature_names(self) -> list:
        return [k for k in self.feature_states.keys()]

    def get_feature_states(self) -> dict:
        return deepcopy(self.feature_states)

    def get_feature_state(self, name: str) -> FeatureState:
        feature_state = self.feature_states.get(name, None)
        return copy(feature_state)

    def set_feature_state(self, feature_state: FeatureState) -> None:
        self.feature_states[feature_state.name] = copy(feature_state)
        self.save(feature_state)



import json
from os import path

from .feature_state import FeatureState
from .state_repository import StateRepository


class FileRepository(StateRepository):
    def __init__(self, filename):
        self.filename = filename
        super().__init__()

    def load(self):
        if not path.isfile(self.filename):
            return {}

        with open(self.filename, 'r') as file:
            data: dict = json.load(file)
            return {k: FeatureState(k, v["enabled"]) for k, v in data.items()}

    def save(self, feature_state):
        data = self.load()
        data[feature_state.name] = {"enabled": feature_state.enabled}
        with open(self.filename, 'w') as file:
            json.dump(data, file)



from .context_locator import ContextLocator
from .feature_context import FeatureContext
from .feature_state import FeatureState
from .state_repository import StateRepository


class Features(object):
    def __init__(self):
        self.global_context = FeatureContext(StateRepository(), None,
                                             snapshot=False,
                                             read_only=False)
        self.context_locator = ContextLocator(self.global_context)

    def configure(self, state_repository: StateRepository = None,
                  snapshot=False, read_only=False):
        self.global_context.configure(state_repository=state_repository,
                                      snapshot=snapshot,
                                      read_only=read_only)

    def __getitem__(self, item) -> bool:
        return self.is_feature_active(item)

    def get_feature_names(self):
        return self.context_locator.peek_context().get_feature_names()

    def set_feature_enabled(self, name: str, enabled: bool = True):
        feature_state = self.get_feature_state(name) or FeatureState(name, enabled)
        feature_state.enabled = enabled
        self.set_feature_state(feature_state)

    def is_feature_active(self, name: str) -> bool:
        return self.context_locator.peek_context().is_feature_active(name)

    def get_feature_state(self, name: str) -> FeatureState:
        return self.context_locator.peek_context().get_feature_state(name)

    def set_feature_state(self, feature_state: FeatureState):
        return self.context_locator.peek_context().set_feature_state(feature_state)

    def new_feature_context(self, state_repository: StateRepository = None,
                            read_only: bool = True,
                            snapshot: bool = True):
        feature_context = self.context_locator.peek_context()
        return FeatureContext(state_repository or feature_context.state_repository,
                              self.context_locator,
                              snapshot=snapshot,
                              read_only=read_only)



import threading


class ContextLocator(object):
    def __init__(self, global_instance):
        self.global_instance = global_instance
        self.thread_local = threading.local()

    def peek_context(self):
        instances = getattr(self.thread_local, "instances", None)
        if instances is not None and len(instances) > 0:
            context = instances[-1]
            if context is not None:
                return context

        return self.global_instance

    def push_context(self, context):
        instances = getattr(self.thread_local, "instances", None)
        if instances is None:
            instances = []
            self.thread_local.instances = instances

        instances.append(context)

    def pop_context(self):
        instances = getattr(self.thread_local, "instances", None)
        if instances is None:
            return

        instances.pop()



from copy import copy

from .context_locator import ContextLocator
from .feature_snapshot import FeatureSnapshot
from .feature_state import FeatureState
from .state_repository import StateRepository


class FeatureContext(object):
    def __init__(self, state_repository: StateRepository,
                 context_locator: ContextLocator = None,
                 snapshot: bool = True,
                 read_only: bool = True):
        self.read_only = read_only
        self.state_repository = state_repository
        self.context_locator = context_locator
        self.configure(state_repository, snapshot, read_only)

    def configure(self, state_repository: StateRepository,
                  snapshot: bool = True,
                  read_only: bool = True):
        self.read_only = read_only
        self.state_repository = state_repository
        if snapshot:
            self.feature_states = FeatureSnapshot(state_repository)
        else:
            self.feature_states = state_repository

    def __getitem__(self, item):
        return self.is_feature_active(item)

    def is_feature_active(self, name):
        feature_state = self.get_feature_state(name)
        return feature_state is not None and feature_state.enabled

    def get_feature_names(self) -> list:
        return self.feature_states.get_feature_names()

    def get_feature_state(self, name: str) -> FeatureState:
        feature_state = self.feature_states.get_feature_state(name)
        return copy(feature_state)

    def set_feature_state(self, feature_state: FeatureState) -> None:
        if self.read_only:
            raise RuntimeError("Feature context is read only.")
        self.feature_states.set_feature_state(copy(feature_state))

    def __enter__(self) -> 'FeatureContext':
        self.context_locator.push_context(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.context_locator.pop_context()



import boto3
from botocore.exceptions import ClientError

from .feature_state import FeatureState
from .state_repository import StateRepository


class DynamodbRepository(StateRepository):
    def __init__(self, dynamodb_resource=None, table_name="features"):
        self.table_name = table_name
        self.dynamodb_resource = dynamodb_resource or boto3.resource("dynamodb")
        super().__init__()

    def load(self):
        return {f.name: f for f in self.__scan_features()}

    def __scan_features(self):
        try:
            table = self.dynamodb_resource.Table(self.table_name)
            response = table.scan()
            items = response.get('Items', [])
            for item in items:
                name = item["name"]
                enabled = item["enabled"]
                yield FeatureState(name, enabled)

            while 'LastEvaluatedKey' in response:
                response = table.scan(ExclusiveStartKey=response['LastEvaluatedKey'])
                items = response.get('Items', [])
                for item in items:
                    name = item["name"]
                    enabled = item["enabled"]
                    yield FeatureState(name, enabled)
        except ClientError as e:
            if e.response["Error"]["Code"] != 'ResourceNotFoundException':
                raise

    def save(self, feature_state):
        table = self.dynamodb_resource.Table(self.table_name)
        table.put_item(Item={
            "name": feature_state.name,
            "enabled": feature_state.enabled
        })
